import typing

# ASGI types.
# Copied from: https://github.com/encode/starlette/blob/master/starlette/types.py
Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]
Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]
ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]



import typing
from contextlib import AsyncExitStack
from types import TracebackType

from ._concurrency import detect_concurrency_backend
from ._exceptions import LifespanNotSupported
from ._types import ASGIApp, Message, Receive, Scope, Send


def state_middleware(app: ASGIApp, state: typing.Dict[str, typing.Any]) -> ASGIApp:
    async def app_with_state(scope: Scope, receive: Receive, send: Send) -> None:
        scope["state"] = state
        await app(scope, receive, send)

    return app_with_state


class LifespanManager:
    def __init__(
        self,
        app: ASGIApp,
        startup_timeout: typing.Optional[float] = 5,
        shutdown_timeout: typing.Optional[float] = 5,
    ) -> None:
        self._state: typing.Dict[str, typing.Any] = {}
        self.app = state_middleware(app, self._state)
        self.startup_timeout = startup_timeout
        self.shutdown_timeout = shutdown_timeout

        self._concurrency_backend = detect_concurrency_backend()
        self._startup_complete = self._concurrency_backend.create_event()
        self._shutdown_complete = self._concurrency_backend.create_event()
        self._receive_queue = self._concurrency_backend.create_queue(capacity=2)
        self._receive_called = False
        self._app_exception: typing.Optional[BaseException] = None
        self._exit_stack = AsyncExitStack()

    async def startup(self) -> None:
        await self._receive_queue.put({"type": "lifespan.startup"})
        await self._concurrency_backend.run_and_fail_after(
            self.startup_timeout, self._startup_complete.wait
        )
        if self._app_exception:
            # Let the caller deal with the exception.
            raise self._app_exception

    async def shutdown(self) -> None:
        await self._receive_queue.put({"type": "lifespan.shutdown"})
        await self._concurrency_backend.run_and_fail_after(
            self.shutdown_timeout, self._shutdown_complete.wait
        )

    async def receive(self) -> Message:
        self._receive_called = True
        return await self._receive_queue.get()

    async def send(self, message: Message) -> None:
        if not self._receive_called:
            raise LifespanNotSupported(
                "Application called send() before receive(). "
                "Is it missing `assert scope['type'] == 'http'` or similar?"
            )

        if message["type"] == "lifespan.startup.complete":
            self._startup_complete.set()
        elif message["type"] == "lifespan.shutdown.complete":
            self._shutdown_complete.set()

    async def run_app(self) -> None:
        scope: Scope = {"type": "lifespan"}

        try:
            await self.app(scope, self.receive, self.send)
        except BaseException as exc:
            self._app_exception = exc

            # We crashed, so don't make '.startup()' and '.shutdown()'
            # wait unnecessarily (or they'll timeout).
            self._startup_complete.set()
            self._shutdown_complete.set()

            if not self._receive_called:
                raise LifespanNotSupported(
                    "Application failed before making its first call to 'receive()'. "
                    "We expect this to originate from a statement similar to "
                    "`assert scope['type'] == 'type'`. "
                    "If that is not the case, then this crash is unexpected and "
                    "there is probably more debug output in the cause traceback."
                ) from exc

            raise

    async def __aenter__(self) -> "LifespanManager":
        await self._exit_stack.__aenter__()
        await self._exit_stack.enter_async_context(
            self._concurrency_backend.run_in_background(self.run_app)
        )
        try:
            await self.startup()
            return self
        except BaseException:
            await self._exit_stack.aclose()
            raise

    async def __aexit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]] = None,
        exc_value: typing.Optional[BaseException] = None,
        traceback: typing.Optional[TracebackType] = None,
    ) -> typing.Optional[bool]:
        if exc_type is None:
            self._exit_stack.push_async_callback(self.shutdown)
        return await self._exit_stack.__aexit__(exc_type, exc_value, traceback)



import asyncio
import contextlib
import types
import typing

from .base import BaseEvent, BaseQueue, ConcurrencyBackend


class AsyncioEvent(BaseEvent):
    def __init__(self) -> None:
        self._event = asyncio.Event()

    def set(self) -> None:
        self._event.set()

    async def wait(self) -> None:
        await self._event.wait()


class AsyncioQueue(BaseQueue):
    def __init__(self, capacity: int) -> None:
        self._queue: asyncio.Queue[typing.Any] = asyncio.Queue(maxsize=capacity)

    async def get(self) -> typing.Any:
        return await self._queue.get()

    async def put(self, value: typing.Any) -> None:
        await self._queue.put(value)


class AsyncioBackend(ConcurrencyBackend):
    def create_event(self) -> BaseEvent:
        return AsyncioEvent()

    def create_queue(self, capacity: int) -> BaseQueue:
        return AsyncioQueue(capacity=capacity)

    async def run_and_fail_after(
        self,
        seconds: typing.Optional[float],
        coroutine: typing.Callable[[], typing.Awaitable[None]],
    ) -> None:
        try:
            await asyncio.wait_for(coroutine(), timeout=seconds)
        except asyncio.TimeoutError:
            raise TimeoutError

    def run_in_background(
        self, coroutine: typing.Callable[[], typing.Awaitable[None]]
    ) -> typing.AsyncContextManager:
        return Background(coroutine)


class Background:
    def __init__(self, coroutine: typing.Callable[[], typing.Awaitable[None]]) -> None:
        self.coroutine = coroutine
        self.task: typing.Optional[asyncio.Task] = None
        self._task_exception: typing.Optional[BaseException] = None

    async def __aenter__(self) -> None:
        async def run_and_silence_cancelled() -> None:
            with contextlib.suppress(asyncio.CancelledError):
                await self.coroutine()

        loop = asyncio.get_event_loop()
        self.task = loop.create_task(run_and_silence_cancelled())

    async def __aexit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]] = None,
        exc_value: typing.Optional[BaseException] = None,
        traceback: typing.Optional[types.TracebackType] = None,
    ) -> None:
        assert self.task is not None

        _, pending = await asyncio.wait({self.task}, timeout=0)
        if pending:
            self.task.cancel()

        await self.task

        if exc_type is None:
            self.task.result()



import types
import typing

import trio

from .._compat import AsyncExitStack
from .base import BaseEvent, BaseQueue, ConcurrencyBackend


class TrioEvent(BaseEvent):
    def __init__(self) -> None:
        self._event = trio.Event()

    def set(self) -> None:
        self._event.set()

    async def wait(self) -> None:
        await self._event.wait()


class TrioQueue(BaseQueue):
    def __init__(self, capacity: int) -> None:
        self._send_channel, self._receive_channel = trio.open_memory_channel(
            max_buffer_size=capacity
        )

    async def get(self) -> typing.Any:
        return await self._receive_channel.receive()

    async def put(self, value: typing.Any) -> None:
        await self._send_channel.send(value)


class TrioBackend(ConcurrencyBackend):
    def create_event(self) -> BaseEvent:
        return TrioEvent()

    def create_queue(self, capacity: int) -> BaseQueue:
        return TrioQueue(capacity=capacity)

    async def run_and_fail_after(
        self,
        seconds: typing.Optional[float],
        coroutine: typing.Callable[[], typing.Awaitable[None]],
    ) -> None:
        with trio.move_on_after(seconds if seconds is not None else float("inf")):
            await coroutine()
            return
        raise TimeoutError

    def run_in_background(
        self, coroutine: typing.Callable[[], typing.Awaitable[None]]
    ) -> typing.AsyncContextManager:
        return Background(coroutine)


class Background:
    def __init__(self, coroutine: typing.Callable[[], typing.Awaitable[None]]) -> None:
        self.coroutine = coroutine
        self._exit_stack = AsyncExitStack()

    async def __aenter__(self) -> None:
        nursery = await self._exit_stack.enter_async_context(trio.open_nursery())
        nursery.start_soon(self.coroutine)

    async def __aexit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]] = None,
        exc_value: typing.Optional[BaseException] = None,
        traceback: typing.Optional[types.TracebackType] = None,
    ) -> None:
        await self._exit_stack.__aexit__(exc_type, exc_value, traceback)



import typing


class BaseEvent:
    def set(self) -> None:
        raise NotImplementedError  # pragma: no cover

    async def wait(self) -> None:
        raise NotImplementedError  # pragma: no cover


class BaseQueue:
    async def get(self) -> typing.Any:
        raise NotImplementedError  # pragma: no cover

    async def put(self, value: typing.Any) -> None:
        raise NotImplementedError  # pragma: no cover


class ConcurrencyBackend:
    def create_event(self) -> BaseEvent:
        raise NotImplementedError  # pragma: no cover

    def create_queue(self, capacity: int) -> BaseQueue:
        raise NotImplementedError  # pragma: no cover

    async def run_and_fail_after(
        self,
        seconds: typing.Optional[float],
        coroutine: typing.Callable[[], typing.Awaitable[None]],
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    def run_in_background(
        self, coroutine: typing.Callable[[], typing.Awaitable[None]]
    ) -> typing.AsyncContextManager:
        raise NotImplementedError  # pragma: no cover



import sniffio

from .base import ConcurrencyBackend


def detect_concurrency_backend() -> ConcurrencyBackend:
    library = sniffio.current_async_library()

    if library == "asyncio":
        from .asyncio import AsyncioBackend

        return AsyncioBackend()
    elif library == "trio":
        from .trio import TrioBackend

        return TrioBackend()

    raise NotImplementedError(
        f"Unsupported async library: {library}"
    )  # pragma: no cover
